###############################################################################
# File:         metric.py
# Description:  Measure the accuracy of the prediction
# Authors:      Luong-Ha Nguyen & James-A. Goulet
# Created:      October 13, 2022
# Updated:      October 21, 2022
# Contact:      luongha.nguyen@gmail.com & james.goulet@polymtl.ca
# Copyright (c) 2022 Luong-Ha Nguyen & James-A. Goulet. Some rights reserved.
###############################################################################
import numpy as np


def mse(prediction: np.ndarray, observation: np.ndarray) -> float:
    """ Mean squared error"""
    return np.nanmean((prediction - observation)**2)


def log_likelihood(prediction: np.ndarray, observation: np.ndarray,
                   std: np.ndarray) -> float:
    """Compute the averaged log-likelihood"""

    log_lik = -0.5 * np.log(2 * np.pi * (std**2)) - 0.5 * ((
        (observation - prediction) / std)**2)

    return np.nanmean(log_lik)


def rmse(prediction: np.ndarray, observation: np.ndarray) -> None:
    """Root mean squared error"""
    mse = mse(prediction, observation)

    return mse**0.5


def classification_error(prediction: np.ndarray, label: np.ndarray) -> None:
    """Compute the classification error"""
    count = 0
    for pred, lab in zip(prediction.T, label):
        if pred != lab:
            count += 1

    return count / len(prediction)



###############################################################################
# File:         tagi_utils.py
# Description:  Python frontend for TAGI utility functions
# Authors:      Luong-Ha Nguyen & James-A. Goulet
# Created:      October 19, 2022
# Updated:      December 04, 2022
# Contact:      luongha.nguyen@gmail.com & james.goulet@polymtl.ca
# Copyright (c) 2022 Luong-Ha Nguyen & James-A. Goulet. Some rights reserved.
###############################################################################
from typing import Tuple, Union

import numpy as np
import pandas as pd
from cutagi import HrSoftmax, UtilityWrapper

from pytagi import Param


class HierarchicalSoftmax(HrSoftmax):
    """Hierarchical softmax wrapper. Further details can be found here
    https://building-babylon.net/2017/08/01/hierarchical-softmax
    """

    def __init__(self) -> None:
        super().__init__()


class Utils:
    """Frontend for utility functions from C++/CUDA backend

    Attributes:
        backend_utils: Utility functionalities from the backend 
    """

    backend_utils = UtilityWrapper()

    def __init__(self) -> None:
        pass

    def label_to_obs(self, labels: np.ndarray,
                     num_classes: int) -> Tuple[np.ndarray, np.ndarray, int]:
        """Get observations and observation indices of the binary tree for
            classification

        Args:
            labels: Labels of dataset
            num_classes: Total number of classes
        Returns:
            obs: Encoded observations of the labels
            obs_idx: Indices of the encoded observations in the output vector
            num_obs: Number of encoded observations
        """

        obs, obs_idx, num_obs = self.backend_utils.label_to_obs_wrapper(
            labels, num_classes)

        return np.array(obs), np.array(obs_idx), np.array(num_obs)

    def load_mnist_images(self, image_file: str, label_file: str,
                          num_images: int) -> Tuple[np.ndarray, np.ndarray]:
        """Load mnist dataset

        Args:
            image_file: Location of the Mnist image file
            label_file: Location of the Mnist label file
            num_images: Number of images to be loaded
        Returns:
            images: Image dataset
            labels: Label dataset
            num_images: Total number of images
        """
        images, labels = self.backend_utils.load_mnist_dataset_wrapper(
            image_file, label_file, num_images)

        return images, labels


    def load_cifar_images(self, image_file: str,
                          num: int) -> Tuple[np.ndarray, np.ndarray]:
        """Load cifar dataset

        Args:
            image_file: Location of image file
            num: Number of images to be loaded
        Returns:
            images: Image dataset
            labels: Label dataset
        """

        images, labels = self.backend_utils.load_cifar_dataset_wrapper(
            image_file, num)

        return images, labels

    def get_labels(self, ma: np.ndarray, Sa: np.ndarray,
                   hr_softmax: HierarchicalSoftmax, num_classes: int,
                   batch_size: int) -> Tuple[np.ndarray, np.ndarray]:
        """Convert last layer's hidden state to labels

        Args:
            ma: Mean of activation units for the output layer
            Sa: Variance of activation units for the output layer
            hr_softmax: Hierarchical softmax
            num_classes: Total number of classes
            batch_size: Number of data in a batch
        Returns:
            pred: Label prediciton
            prob: Probability for each label
        """

        pred, prob = self.backend_utils.get_labels_wrapper(
            ma, Sa, hr_softmax, num_classes, batch_size)

        return pred, prob

    def get_errors(self, ma: np.ndarray, Sa: np.ndarray, labels: np.ndarray,
                   hr_softmax: HierarchicalSoftmax, num_classes: int,
                   batch_size: int) -> Tuple[np.ndarray, np.ndarray]:
        """Convert last layer's hidden state to labels

        Args:
            ma: Mean of activation units for the output layer
            Sa: Variance of activation units for the output layer
            labels: Label dataset
            hr_softmax: Hierarchical softmax
            num_classes: Total number of classes
            batch_size: Number of data in a batch
        Returns:
            pred: Label prediction
            prob: Probability for each label
        """

        pred, prob = self.backend_utils.get_error_wrapper(
            ma, Sa, labels, hr_softmax, num_classes, batch_size)

        return pred, prob

    def get_hierarchical_softmax(self,
                                 num_classes: int) -> HierarchicalSoftmax:
        """Convert labels to binary tree

        Args:
            num_classes: Total number of classes
        Returns:
            hr_softmax: Hierarchical softmax
        """
        hr_softmax = self.backend_utils.hierarchical_softmax_wrapper(
            num_classes)

        return hr_softmax

    def obs_to_label_prob(self, ma: np.ndarray, Sa: np.ndarray,
                          hr_softmax: HierarchicalSoftmax,
                          num_classes: int) -> np.ndarray:
        """Convert observation to label probabilities

        Args:
            ma: Mean of activation units for the output layer
            Sa: Variance of activation units for the output layer
            hr_softmax: Hierarchical softmax
            num_classes: Total number of classes
        Returns:
            prob: Probability for each label
        """

        prob = self.backend_utils.obs_to_label_prob_wrapper(
            ma, Sa, hr_softmax, num_classes)

        return np.array(prob)

    def create_rolling_window(self, data: np.ndarray, output_col: np.ndarray,
                              input_seq_len: int, output_seq_len: int,
                              num_features: int,
                              stride: int) -> Tuple[np.ndarray, np.ndarray]:
        """Create rolling window for time series data

        Args: 
            data: dataset
            output_col: Indices of the output columns
            input_seq_len: Length of the input sequence
            output_seq_len: Length of the output sequence
            num_features: Number of features 
            stride: Controls number of steps for the window movements
        Returns: 
            input_data: Input data for neural networks in sequence
            output_data: Output data for neural networks in sequence
        """
        num_data = int((len(data) / num_features - input_seq_len -
                        output_seq_len) / stride + 1)

        input_data, output_data = self.backend_utils.create_rolling_window_wrapper(
            data.flatten(), output_col, input_seq_len, output_seq_len,
            num_features, stride)
        input_data = input_data.reshape((num_data, input_seq_len))
        output_data = output_data.reshape((num_data, output_seq_len))

        return input_data, output_data

    def get_upper_triu_cov(self, batch_size: int, num_data: int,
                           sigma: float) -> np.ndarray:
        """Create an upper triangle covriance matrix for inputs"""

        vx_f = self.backend_utils.get_upper_triu_cov_wrapper(
            batch_size, num_data, sigma)

        return np.array(vx_f)


def exponential_scheduler(curr_v: float, min_v: float, decaying_factor: float,
                          curr_iter: float) -> float:
    """Exponentially decaying"""

    return np.maximum(curr_v * (decaying_factor**curr_iter), min_v)


class Normalizer:
    """Different method to normalize the data before feeding
    to neural networks"""

    def __init__(self, method: Union[str, None] = None) -> None:
        self.method = method

    def standardize(self, data: np.ndarray, mu: np.ndarray,
                    std: np.ndarray) -> np.ndarray:
        """Z-score normalization where 
        data_norm = (data - data_mean) / data_std """

        return (data - mu) / (std + 1e-10)

    @staticmethod
    def unstandardize(norm_data: np.ndarray, mu: np.ndarray,
                      std: np.ndarray) -> np.ndarray:
        """Transform standardized data to original space"""
        return norm_data * (std + 1e-10) + mu

    @staticmethod
    def unstandardize_std(norm_std: np.ndarray, std: np.ndarray) -> np.ndarray:
        """Transform standardized std to original space"""

        return norm_std * (std + 1e-10)

    def max_min_norm(self, data: np.ndarray, max_value: np.ndarray,
                     min_value: np.ndarray) -> np.ndarray:
        """Normalize the data between 0 and 1"""
        assert np.all(max_value > min_value)
        return (data - min_value) / (max_value - min_value + 1e-10)

    @staticmethod
    def max_min_unnorm(norm_data: np.ndarray, max_value: np.ndarray,
                       min_value: np.ndarray) -> np.ndarray:
        """Transform max-min normalized data to original space"""

        return (norm_data * (max_value - min_value + 1e-10)) + min_value

    @staticmethod
    def max_min_unnorm_std(norm_std: np.ndarray, max_value: np.ndarray,
                           min_value: np.ndarray) -> np.ndarray:
        """Transform max-min normalized std to original space"""

        return (norm_std * (max_value - min_value + 1e-10))

    @staticmethod
    def compute_mean_std(data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Compute sample mean and standard deviation"""

        return (np.nanmean(data, axis=0), np.nanstd(data, axis=0))

    @staticmethod
    def compute_max_min(data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Compute max min values"""

        return (np.nanmax(data, axis=0), np.nanmin(data, axis=0))


def load_param_from_files(mw_file: str, Sw_file: str, mb_file: str,
                          Sb_file: str, mw_sc_file: str, Sw_sc_file: str,
                          mb_sc_file: str, Sb_sc_file: str) -> Param:
    """Load parameter from csv file"""
    mw_df = pd.read_csv(mw_file, header=None)
    Sw_df = pd.read_csv(Sw_file, header=None)
    mb_df = pd.read_csv(mb_file, header=None)
    Sb_df = pd.read_csv(Sb_file, header=None)

    try:
        mw_sc_df = pd.read_csv(mw_sc_file, header=None)
        Sw_sc_df = pd.read_csv(Sw_sc_file, header=None)
        mb_sc_df = pd.read_csv(mb_sc_file, header=None)
        Sb_sc_df = pd.read_csv(Sb_sc_file, header=None)
    except ValueError:
        mw_sc_df = pd.DataFrame()
        Sw_sc_df = pd.DataFrame()
        mb_sc_df = pd.DataFrame()
        Sb_sc_df = pd.DataFrame()

    return Param(mw=mw_df.values,
                 Sw=Sw_df.values,
                 mb=mb_df.values,
                 Sb=Sb_df.values,
                 mw_sc=mw_sc_df.values,
                 Sw_sc=Sw_sc_df.values,
                 mb_sc=mb_sc_df.values,
                 Sb_sc=Sb_sc_df.values)



###############################################################################
# File:         tagi_network.py
# Description:  Python frontend for TAGI network
# Authors:      Luong-Ha Nguyen & James-A. Goulet
# Created:      October 13, 2022
# Updated:      December 11, 2022
# Contact:      luongha.nguyen@gmail.com & james.goulet@polymtl.ca
# Copyright (c) 2022 Luong-Ha Nguyen & James-A. Goulet. Some rights reserved.
###############################################################################
from typing import Tuple

import numpy as np
import cutagi as tagi


class NetProp(tagi.Network):
    """Base class for network properties defined in the backend C++/CUDA
    Layer code:
        1: Fully-connected layer
        2: Convolutional layer
        21: Transpose convolutional layer
        3: Max pooling layer (currently not supported)
        4: Average pooling
        5: Layer normalization
        6: Batch normalization
        7: LSTM layer

    Activation code:
        0: No activation 
        1: Tanh
        2: Sigmoid
        4: ReLU
        5: Softplus
        6: Leakyrelu
        7: Mixture ReLU
        8: Mixture bounded ReLU i.e., [-1.0, 1.0]

    Attributes:
        layers: A vector contains different layers of network architecture
        nodes: Number of hidden units
        kernels: Kernel size fo convolutional layer
        widths: Width of image
        heights: Heights of image
        filters: Number of filters i.e. depth of image for each layer
        activations: Activation function
        pads: Padding that applied to image
        pad_types: Type of padding
        shortcuts: Layer index for residual network 
        mu_v2b: Mean of the observation noise squared
        sigma_v2b: Standard deviation of the observation noise squared
        sigma_v: Observation noise
        decay_factor_sigma_v: Decaying factor for sigma v (default value: 0.99)
        sigma_v_min: Minimum value of observation noise (default value: 0.3)
        sigma_x: Input noise noise
        is_output_ud: Whether or not to update output layer
        is_idx_ud: Wheher or not to update only hidden units in the output
                   layers
        last_backward_layer: Index of last layer whose hidden states are updated
        nye: Number of observation for hierarchical softmax
        noise_gain : Gain fof biases parameters relating to noise's hidden
            states
        noise_type: homosce or heteros
        batch_size: Number of batches of data
        input_seq_len: Sequence lenth for lstm inputs
        input_seq_len: Sequence lenth for last layer's outputs
        seq_stride: Spacing between sequences for lstm layer
        multithreading: Whether or not to run parallel computing using multiple
            threads
        collect_derivative: Enable the derivative computation mode
        is_full_cov: Enable full covariance mode
        init_method: Initalization method e.g. He and Xavier
        device: Either cpu or cuda
        ra_mt: Momentum for the normalization layer
    """
    layers: list
    nodes: list
    kernels: list
    strides: list
    widths: list
    heights: list
    filters: list
    activation: list
    pads: list
    pad_types: list
    shortcuts: list
    mu_v2b: np.ndarray
    sigma_v2b: np.ndarray
    sigma_v: float
    decay_factor_sigma_v: float
    sigma_v_min: float
    sigma_x: float
    is_idx_ud: bool
    is_output_ud: bool
    last_backward_layer: int
    nye: int
    noise_gain: float
    noise_type: str
    batch_size: int
    input_seq_len: int
    output_seq_len: int
    seq_stride: int
    multithreading: bool
    collect_derivative: bool
    is_full_cov: bool
    init_method: str
    device: str
    ra_mt: float

    def __init__(self) -> None:
        super().__init__()


class Param(tagi.Param):
    """Frontend apt for weight and biases

    Attributes:
        mw: Mean of weight parameters
        Sw: Variance of weight parameters
        mb: Mean of bias parameters
        Sb: Variance of bias parameters
        mw_sc: Mean of weight parameters for the residual network
        Sw_sc: Variance of weight parameters for the residual network
        mb_sc: Mean of bias parameters for the residual network
        Sb_sc: Variance of bias parameters for the residual network
    """
    mw: np.ndarray
    Sw: np.ndarray
    mb: np.ndarray
    Sb: np.ndarray
    mw_sc: np.ndarray
    Sw_sc: np.ndarray
    mb_sc: np.ndarray
    Sb_sc: np.ndarray

    def __init__(self, mw: np.ndarray, Sw: np.ndarray, mb: np.ndarray,
                 Sb: np.ndarray, mw_sc: np.ndarray, Sw_sc: np.ndarray,
                 mb_sc: np.ndarray, Sb_sc: np.ndarray) -> None:
        super().__init__()
        self.mw = mw
        self.Sw = Sw
        self.mb = mb
        self.Sb = Sb
        self.mw_sc = mw_sc
        self.Sw_sc = Sw_sc
        self.mb_sc = mb_sc
        self.Sb_sc = Sb_sc


class TagiNetwork:
    """Python frontend calling TAGI network in C++/CUDA backend

    Attributes:
        network: Network wrapper that calls the tagi network from
         the backend 
        net_prop: Network properties
    """

    network: tagi.NetworkWrapper

    def __init__(self, net_prop: NetProp) -> None:
        self.net_prop = net_prop

    @property
    def net_prop(self) -> NetProp():
        """"Get network properties"""
        return self._net_prop

    @net_prop.setter
    def net_prop(self, value: NetProp) -> None:
        """Set network properties"""
        self._net_prop = value
        self.network = tagi.NetworkWrapper(self._net_prop)

    def feed_forward(self, x_batch: np.ndarray, Sx_batch: np.ndarray,
                     Sx_f_batch: np.ndarray) -> None:
        """Forward pass
        the size of x_batch, Sx_batch (B, N) where B is the batch size and N
        is the data dimension

        Args:
            x_batch: Input data
            Sx_batch: Diagonal variance of input data
            Sx_f_batch: Full variance of input data
        """
        self.network.feed_forward_wrapper(x_batch.flatten(),
                                          Sx_batch.flatten(),
                                          Sx_f_batch.flatten())

    def connected_feed_forward(self, ma: np.ndarray, va: np.ndarray,
                               mz: np.ndarray, vz: np.ndarray,
                               jcb: np.ndarray) -> None:
        """Forward pass for the network that is connected to the other 
        network e.g., decoder network in autoencoder task where its inputs 
        are the outputs of the encoder network.

        Args:
            ma: Mean of activation units
            va: Variance of activation units
            mz: Mean of hidden states
            vz: Variance of hidden states
            jcb: Jacobian matrix (da/dz)
        """

        self.network.connected_feed_forward_wrapper(ma, va, mz, vz, jcb)

    def state_feed_backward(self, y_batch: np.ndarray, v_batch: np.ndarray,
                            ud_idx_batch: np.ndarray) -> None:
        """Update hidden states
        the size of y_batch, V_batch (B, N) where B is the batch size and N
        is the data dimension

        Args:
            y_batch: Observations 
            v_batch: Variance of observations
            ud_idx_batch: Updated indices for the last layer e.g., it 
                only required for classification task
        """
        self.network.state_feed_backward_wrapper(y_batch.flatten(),
                                                 v_batch.flatten(),
                                                 ud_idx_batch.flatten())

    def param_feed_backward(self) -> None:
        """Update parameters"""
        self.network.param_feed_backward_wrapper()

    def get_network_outputs(self) -> Tuple[np.ndarray, np.ndarray]:
        """Get output layer's hidden state distribution
        
        Returns:
            ma: Mean of activation units
            va: Variance of activation units
        """
        ma, va = self.network.get_network_outputs_wrapper()

        return ma, va

    def get_network_predictions(self) -> Tuple[np.ndarray, np.ndarray]:
        """Get distribution of the predictions
        
        Returns:
            m_pred: Mean of predictions
            v_pred: Variance of predictions
        """
        m_pred, v_pred = self.network.get_network_prediction_wrapper()

        return m_pred, v_pred

    def get_all_network_outputs(
        self
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Get all hidden states of the output layers

        Returns:
            ma: Mean of activations for the output layer
            va: Variance of activations for the output layer
            mz: Mean of hidden states for the output layer
            vz: Variance of hidden states for the output layer       
            jcb: Jacobian matrix for the output layer
        """
        ma, va, mz, vz, jcb = self.network.get_all_network_outputs_wrapper()

        return (ma, va, mz, vz, jcb)

    def get_all_network_inputs(
        self
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Get all hidden states of the output layers

        Returns:
            ma: Mean of activations for the input layer
            va: Variance of activations for the input layer
            mz: Mean of hidden states for the input layer
            vz: Variance of hidden states for the input layer     
            jcb: Jacobian matrix for the input layer
        """
        ma, va, mz, vz, jcb = self.network.get_all_network_inputs_wrapper()

        return (ma, va, mz, vz, jcb)

    def get_derivatives(self, layer: int = 0) -> Tuple[np.ndarray, np.ndarray]:
        """ Compute derivatives of the output layer w.r.t a given layer using TAGI
        
        Args:
            layer: Layer index of the network
        Returns:
            mdy: Mean values of derivatives
            vdy: Variance values of derivatives
        """
        mdy, vdy = self.network.get_derivative_wrapper(layer)
        return mdy, vdy

    def get_inovation_mean_var(self,
                               layer: int) -> Tuple[np.ndarray, np.ndarray]:
        """Get updating quantities for the inovation

        Args: 
            layer: Layer index of the network
        
        Returns:
            delta_m: Inovation mean 
            delta_v: Inovation variance 
        """

        delta_m, delta_v = self.network.get_inovation_mean_var_wrapper(layer)

        return delta_m, delta_v

    def get_state_delta_mean_var(self) -> None:
        """Get updating quatities for the first layer
        
        Returns:
            delta_mz: Updating quantities for the hidden-state mean of the
                first layer
            delta_vz: Updating quantities for the hidden-state variance of the
                first layer
        """
        delta_mz, delta_vz = self.network.get_state_delta_mean_var_wrapper()

        return delta_mz, delta_vz

    def set_parameters(self, param: Param) -> None:
        """Set parameter values to network"""
        self.network.set_parameters_wrapper(param)

    def get_parameters(self) -> tagi.Param:
        """Get parameters of network"""
        return self.network.get_parameters_wrapper()
